"""Test matrix-function-vector products via Lanczos' algorithm."""

from matfree import lanczos, test_util
from matfree.backend import linalg, np, prng


def test_funm_vector_product(n=11):
    """Test matrix-function-vector products via Lanczos' algorithm."""
    # Create a test-problem: matvec, matrix function,
    # vector, and parameters (a matrix).

    def matvec(x, p):
        return p @ x

    # todo: write a test for matfun=np.inv,
    #  because this application seems to be brittle
    def fun(x):
        return np.sin(x)

    v = prng.normal(prng.prng_key(2), shape=(n,))

    eigvals = np.linspace(0.01, 0.99, num=n)
    matrix = test_util.symmetric_matrix_from_eigenvalues(eigvals)

    # Compute the solution
    eigvals, eigvecs = linalg.eigh(matrix)
    log_matrix = eigvecs @ linalg.diagonal(fun(eigvals)) @ eigvecs.T
    expected = log_matrix @ v

    # Compute the matrix-function vector product
    order = 6
    matfun_vec = lanczos.funm_vector_product_spd(fun, order, matvec)
    received = matfun_vec(v, matrix)
    assert np.allclose(expected, received, atol=1e-6)



"""Approximate matrix-function-vector products with polynomial expansions.

This module does not include Lanczos-style approximations, which are
in [matfree.lanczos][matfree.lanczos].
"""

from matfree.backend import containers, control_flow, np
from matfree.backend.typing import Array


def funm_vector_product_chebyshev(matfun, order, matvec, /):
    """Compute a matrix-function-vector product via Chebyshev's algorithm.

    This function assumes that the **spectrum of the matrix-vector product
    is contained in the interval (-1, 1)**, and that the **matrix-function
    is analytic on this interval**. If this is not the case,
    transform the matrix-vector product and the matrix-function accordingly.
    """
    # Construct nodes
    nodes = _chebyshev_nodes(order)
    fx_nodes = matfun(nodes)

    class _ChebyshevState(containers.NamedTuple):
        interpolation: Array
        poly_coefficients: tuple[Array, Array]
        poly_values: tuple[Array, Array]

    def init_func(vec, *parameters):
        # Initialize the scalar recursion
        # (needed to compute the interpolation weights)
        t2_n, t1_n = nodes, np.ones_like(nodes)
        c1 = np.mean(fx_nodes * t1_n)
        c2 = 2 * np.mean(fx_nodes * t2_n)

        # Initialize the vector-valued recursion
        # (this is where the matvec happens)
        t2_x, t1_x = matvec(vec, *parameters), vec
        value = c1 * t1_x + c2 * t2_x
        return _ChebyshevState(value, (t2_n, t1_n), (t2_x, t1_x))

    def recursion_func(val: _ChebyshevState, *parameters) -> _ChebyshevState:
        value, (t2_n, t1_n), (t2_x, t1_x) = val

        # Apply the next scalar recursion and
        # compute the next coefficient
        t2_n, t1_n = 2 * nodes * t2_n - t1_n, t2_n
        c2 = 2 * np.mean(fx_nodes * t2_n)

        # Apply the next matrix-vector product recursion and
        # compute the next interpolation-value
        t2_x, t1_x = 2 * matvec(t2_x, *parameters) - t1_x, t2_x
        value += c2 * t2_x
        return _ChebyshevState(value, (t2_n, t1_n), (t2_x, t1_x))

    def extract_func(val: _ChebyshevState):
        return val.interpolation

    alg = (0, order - 1), init_func, recursion_func, extract_func
    return _funm_vector_product_polyexpand(alg)


def _chebyshev_nodes(n, /):
    k = np.arange(n, step=1.0) + 1
    return np.cos((2 * k - 1) / (2 * n) * np.pi())


def _funm_vector_product_polyexpand(matrix_poly_alg, /):
    """Implement a matrix-function-vector product via a polynomial expansion."""
    (lower, upper), init_func, step_func, extract_func = matrix_poly_alg

    def matvec(vec, *parameters):
        final_state = control_flow.fori_loop(
            lower=lower,
            upper=upper,
            body_fun=lambda _i, v: step_func(v, *parameters),
            init_val=init_func(vec, *parameters),
        )
        return extract_func(final_state)

    return matvec



"""Pseudo-inverses of linear operators."""

from matfree.backend.typing import Callable

# todo: what happens if we want to reuse e.g. LU decompositions of AA^*?
# todo: should we not just return pinv-matvec but also pinv-vecmat?


def pinv_tall(Av: Callable, vA: Callable, *, solve: Callable) -> Callable:
    r"""Construct the (Moore-Penrose) pseudo-inverse of a full-rank, ''tall'' matrix.

    Implemented as a left-inverse, $A^\dagger := A^* (A A^*)^{-1}$,
    which is applicable if $A$ has full **column**-rank.

    Parameters
    ----------
    Av:
        Matrix-vector product function.
    vA:
        Vector-matrix product function.
    solve:
        Solution of a linear system. Maps a matrix-vector production function and a
        right-hand-side vector to the solution of the linear system.

    Returns
    -------
    Callable
        Matrix-vector product function that implements
        the pseudo-inverse of the matrix-vector product function `Av`.
    """

    def pinv(s):
        return vA(solve(lambda v: Av(vA(v)), s))

    return pinv


def pinv_wide(Av: Callable, vA: Callable, *, solve: Callable) -> Callable:
    r"""Construct the (Moore-Penrose) pseudo-inverse of a full-rank, ''wide'' matrix.

    Implemented as a right-inverse, $A^\dagger := (A^* A)^{-1} A^*$,
    which is applicable if $A$ has full **row**-rank.

    Parameters
    ----------
    Av:
        Matrix-vector product function.
    vA:
        Vector-matrix product function.
    solve:
        Solution of a linear system. Maps a matrix-vector production function and a
        right-hand-side vector to the solution of the linear system.

    Returns
    -------
    Callable
        Matrix-vector product function that implements
        the pseudo-inverse of the matrix-vector product function `Av`.
    """

    def pinv(s):
        return solve(lambda v: vA(Av(v)), vA(s))

    return pinv



"""All things Lanczos' algorithm.

This includes
stochastic Lanczos quadrature (extending the integrands
in [hutchinson][matfree.hutchinson] to those that implement
stochastic Lanczos quadrature),
Lanczos-implementations of matrix-function-vector products,
and various Lanczos-decompositions of matrices.

Examples
--------
>>> import jax.random
>>> import jax.numpy as jnp
>>>
>>> jnp.set_printoptions(1)
>>>
>>> M = jax.random.normal(jax.random.PRNGKey(1), shape=(10, 10))
>>> A = M.T @ M
>>> v = jax.random.normal(jax.random.PRNGKey(2), shape=(10,))
>>>
>>> # Compute a matrix-logarithm with Lanczos' algorithm
>>> matfun_vec = funm_vector_product_spd(jnp.log, 4, lambda s: A @ s)
>>> matfun_vec(v)
Array([-4. , -2.1, -2.7, -1.9, -1.3, -3.5, -0.5, -0.1,  0.3,  1.5],      dtype=float32)
"""

from matfree.backend import containers, control_flow, func, linalg, np, tree_util
from matfree.backend.typing import Array, Callable, Tuple


def integrand_spd_logdet(order, matvec, /):
    """Construct the integrand for the log-determinant.

    This function assumes a symmetric, positive definite matrix.
    """
    return integrand_spd(np.log, order, matvec)


def integrand_spd(matfun, order, matvec, /):
    """Quadratic form for stochastic Lanczos quadrature.

    This function assumes a symmetric, positive definite matrix.
    """

    def quadform(v0, *parameters):
        v0_flat, v_unflatten = tree_util.ravel_pytree(v0)
        length = linalg.vector_norm(v0_flat)
        v0_flat /= length

        def matvec_flat(v_flat, *p):
            v = v_unflatten(v_flat)
            Av = matvec(v, *p)
            flat, unflatten = tree_util.ravel_pytree(Av)
            return flat

        algorithm = alg_tridiag_full_reortho(matvec_flat, order)
        _, (diag, off_diag) = algorithm(v0_flat, *parameters)
        eigvals, eigvecs = _eigh_tridiag(diag, off_diag)

        # Since Q orthogonal (orthonormal) to v0, Q v = Q[0],
        # and therefore (Q v)^T f(D) (Qv) = Q[0] * f(diag) * Q[0]
        fx_eigvals = func.vmap(matfun)(eigvals)
        return length**2 * linalg.vecdot(eigvecs[0, :], fx_eigvals * eigvecs[0, :])

    return quadform


def integrand_product_logdet(depth, matvec, vecmat, /):
    r"""Construct the integrand for the log-determinant of a matrix-product.

    Here, "product" refers to $X = A^\top A$.
    """
    return integrand_product(np.log, depth, matvec, vecmat)


def integrand_product_schatten_norm(power, depth, matvec, vecmat, /):
    r"""Construct the integrand for the p-th power of the Schatten-p norm."""

    def matfun(x):
        """Matrix-function for Schatten-p norms."""
        return x ** (power / 2)

    return integrand_product(matfun, depth, matvec, vecmat)


def integrand_product(matfun, depth, matvec, vecmat, /):
    r"""Construct the integrand for the trace of a function of a matrix-product.

    Instead of the trace of a function of a matrix,
    compute the trace of a function of the product of matrices.
    Here, "product" refers to $X = A^\top A$.
    """

    def quadform(v0, *parameters):
        v0_flat, v_unflatten = tree_util.ravel_pytree(v0)
        length = linalg.vector_norm(v0_flat)
        v0_flat /= length

        def matvec_flat(v_flat, *p):
            v = v_unflatten(v_flat)
            Av = matvec(v, *p)
            flat, unflatten = tree_util.ravel_pytree(Av)
            return flat, tree_util.partial_pytree(unflatten)

        w0_flat, w_unflatten = func.eval_shape(matvec_flat, v0_flat)
        matrix_shape = (*np.shape(w0_flat), *np.shape(v0_flat))

        def vecmat_flat(w_flat):
            w = w_unflatten(w_flat)
            wA = vecmat(w, *parameters)
            return tree_util.ravel_pytree(wA)[0]

        # Decompose into orthogonal-bidiag-orthogonal
        algorithm = alg_bidiag_full_reortho(
            lambda v: matvec_flat(v)[0], vecmat_flat, depth, matrix_shape=matrix_shape
        )
        output = algorithm(v0_flat, *parameters)
        u, (d, e), vt, _ = output

        # Compute SVD of factorisation
        B = _bidiagonal_dense(d, e)
        _, S, Vt = linalg.svd(B, full_matrices=False)

        # Since Q orthogonal (orthonormal) to v0, Q v = Q[0],
        # and therefore (Q v)^T f(D) (Qv) = Q[0] * f(diag) * Q[0]
        eigvals, eigvecs = S**2, Vt.T
        fx_eigvals = func.vmap(matfun)(eigvals)
        return length**2 * linalg.vecdot(eigvecs[0, :], fx_eigvals * eigvecs[0, :])

    return quadform


def _bidiagonal_dense(d, e):
    diag = linalg.diagonal_matrix(d)
    offdiag = linalg.diagonal_matrix(e, 1)
    return diag + offdiag


def funm_vector_product_spd(matfun, order, matvec, /):
    """Implement a matrix-function-vector product via Lanczos' algorithm.

    This algorithm uses Lanczos' tridiagonalisation with full re-orthogonalisation
    and therefore applies only to symmetric, positive definite matrices.
    """
    algorithm = alg_tridiag_full_reortho(matvec, order)

    def estimate(vec, *parameters):
        length = linalg.vector_norm(vec)
        vec /= length
        basis, (diag, off_diag) = algorithm(vec, *parameters)
        eigvals, eigvecs = _eigh_tridiag(diag, off_diag)

        fx_eigvals = func.vmap(matfun)(eigvals)
        return length * (basis.T @ (eigvecs @ (fx_eigvals * eigvecs[0, :])))

    return estimate


def _eigh_tridiag(diag, off_diag):
    # todo: once jax supports eigh_tridiagonal(eigvals_only=False),
    #  use it here. Until then: an eigen-decomposition of size (order + 1)
    #  does not hurt too much...
    diag = linalg.diagonal_matrix(diag)
    offdiag1 = linalg.diagonal_matrix(off_diag, -1)
    offdiag2 = linalg.diagonal_matrix(off_diag, 1)
    dense_matrix = diag + offdiag1 + offdiag2
    eigvals, eigvecs = linalg.eigh(dense_matrix)
    return eigvals, eigvecs


def svd_approx(
    v0: Array, depth: int, Av: Callable, vA: Callable, matrix_shape: Tuple[int, ...]
):
    """Approximate singular value decomposition.

    Uses GKL with full reorthogonalisation to bi-diagonalise the target matrix
    and computes the full SVD of the (small) bidiagonal matrix.

    Parameters
    ----------
    v0:
        Initial vector for Golub-Kahan-Lanczos bidiagonalisation.
    depth:
        Depth of the Krylov space constructed by Golub-Kahan-Lanczos bidiagonalisation.
        Choosing `depth = min(nrows, ncols) - 1` would yield behaviour similar to
        e.g. `np.linalg.svd`.
    Av:
        Matrix-vector product function.
    vA:
        Vector-matrix product function.
    matrix_shape:
        Shape of the matrix involved in matrix-vector and vector-matrix products.
    """
    # Factorise the matrix
    algorithm = alg_bidiag_full_reortho(Av, vA, depth, matrix_shape=matrix_shape)
    u, (d, e), vt, _ = algorithm(v0)

    # Compute SVD of factorisation
    B = _bidiagonal_dense(d, e)
    U, S, Vt = linalg.svd(B, full_matrices=False)

    # Combine orthogonal transformations
    return u @ U, S, Vt @ vt


class _LanczosAlg(containers.NamedTuple):
    """Lanczos decomposition algorithm."""

    init: Callable
    """Initialise the state of the algorithm. Usually, this involves pre-allocation."""

    step: Callable
    """Compute the next iteration."""

    extract: Callable
    """Extract the solution from the state of the algorithm."""

    lower_upper: Tuple[int, int]
    """Range of the for-loop used to decompose a matrix."""


def alg_tridiag_full_reortho(
    Av: Callable, depth, /, validate_unit_2_norm=False
) -> Callable:
    """Construct an implementation of **tridiagonalisation**.

    Uses pre-allocation. Fully reorthogonalise vectors at every step.

    This algorithm assumes a **symmetric matrix**.

    Decompose a matrix into a product of orthogonal-**tridiagonal**-orthogonal matrices.
    Use this algorithm for approximate **eigenvalue** decompositions.

    """

    class State(containers.NamedTuple):
        i: int
        basis: Array
        tridiag: Tuple[Array, Array]
        q: Array
        length: Array

    def init(init_vec: Array) -> State:
        (ncols,) = np.shape(init_vec)
        if depth >= ncols or depth < 1:
            raise ValueError

        if validate_unit_2_norm:
            init_vec = _validate_unit_2_norm(init_vec)

        diag = np.zeros((depth + 1,))
        offdiag = np.zeros((depth,))
        basis = np.zeros((depth + 1, ncols))

        return State(0, basis, (diag, offdiag), init_vec, 1.0)

    def apply(state: State, *parameters) -> State:
        i, basis, (diag, offdiag), vec, length = state

        # Compute the next off-diagonal entry
        offdiag = offdiag.at[i - 1].set(length)

        # Re-orthogonalise against ALL basis elements before storing.
        # Note: we re-orthogonalise against ALL columns of Q, not just
        # the ones we have already computed. This increases the complexity
        # of the whole iteration from n(n+1)/2 to n^2, but has the advantage
        # that the whole computation has static bounds (thus we can JIT it all).
        # Since 'Q' is padded with zeros, the numerical values are identical
        # between both modes of computing.
        vec, *_ = _gram_schmidt_classical(vec, basis)

        # Store new basis element
        basis = basis.at[i, :].set(vec)

        # When i==0, Q[i-1] is Q[-1] and again, we benefit from the fact
        #  that Q is initialised with zeros.
        vec = Av(vec, *parameters)
        basis_vectors_previous = np.asarray([basis[i], basis[i - 1]])
        vec, length, (coeff, _) = _gram_schmidt_classical(vec, basis_vectors_previous)
        diag = diag.at[i].set(coeff)

        return State(i + 1, basis, (diag, offdiag), vec, length)

    def extract(state: State, /):
        # todo: return final output "_ignored"
        _, basis, (diag, offdiag), *_ignored = state
        return basis, (diag, offdiag)

    alg = _LanczosAlg(
        init=init, step=apply, extract=extract, lower_upper=(0, depth + 1)
    )
    return func.partial(_decompose_fori_loop, algorithm=alg)


def alg_bidiag_full_reortho(
    Av: Callable, vA: Callable, depth, /, matrix_shape, validate_unit_2_norm=False
):
    """Construct an implementation of **bidiagonalisation**.

    Uses pre-allocation. Fully reorthogonalise vectors at every step.

    Works for **arbitrary matrices**. No symmetry required.

    Decompose a matrix into a product of orthogonal-**bidiagonal**-orthogonal matrices.
    Use this algorithm for approximate **singular value** decompositions.
    """
    nrows, ncols = matrix_shape
    max_depth = min(nrows, ncols) - 1
    if depth > max_depth or depth < 0:
        msg1 = f"Depth {depth} exceeds the matrix' dimensions. "
        msg2 = f"Expected: 0 <= depth <= min(nrows, ncols) - 1 = {max_depth} "
        msg3 = f"for a matrix with shape {matrix_shape}."
        raise ValueError(msg1 + msg2 + msg3)

    class State(containers.NamedTuple):
        i: int
        Us: Array
        Vs: Array
        alphas: Array
        betas: Array
        beta: Array
        vk: Array

    def init(init_vec: Array) -> State:
        if validate_unit_2_norm:
            init_vec = _validate_unit_2_norm(init_vec)

        alphas = np.zeros((depth + 1,))
        betas = np.zeros((depth + 1,))
        Us = np.zeros((depth + 1, nrows))
        Vs = np.zeros((depth + 1, ncols))
        v0, _ = _normalise(init_vec)
        return State(0, Us, Vs, alphas, betas, np.zeros(()), v0)

    def apply(state: State, *parameters) -> State:
        i, Us, Vs, alphas, betas, beta, vk = state
        Vs = Vs.at[i].set(vk)
        betas = betas.at[i].set(beta)

        uk = Av(vk, *parameters) - beta * Us[i - 1]
        uk, alpha = _normalise(uk)
        uk, *_ = _gram_schmidt_classical(uk, Us)  # full reorthogonalisation
        Us = Us.at[i].set(uk)
        alphas = alphas.at[i].set(alpha)

        vk = vA(uk, *parameters) - alpha * vk
        vk, beta = _normalise(vk)
        vk, *_ = _gram_schmidt_classical(vk, Vs)  # full reorthogonalisation

        return State(i + 1, Us, Vs, alphas, betas, beta, vk)

    def extract(state: State, /):
        _, uk_all, vk_all, alphas, betas, beta, vk = state
        return uk_all.T, (alphas, betas[1:]), vk_all, (beta, vk)

    alg = _LanczosAlg(
        init=init, step=apply, extract=extract, lower_upper=(0, depth + 1)
    )
    return func.partial(_decompose_fori_loop, algorithm=alg)


def _validate_unit_2_norm(v, /):
    # Lanczos assumes a unit-2-norm vector as an input
    # We cannot raise an error based on values of the init_vec,
    # but we can make it obvious that the result is unusable.
    is_not_normalized = np.abs(linalg.vector_norm(v) - 1.0) > 10 * np.finfo_eps(v.dtype)
    return control_flow.cond(
        is_not_normalized, lambda s: np.nan() * np.ones_like(s), lambda s: s, v
    )


def _gram_schmidt_classical(vec, vectors):  # Gram-Schmidt
    vec, coeffs = control_flow.scan(_gram_schmidt_classical_step, vec, xs=vectors)
    vec, length = _normalise(vec)
    return vec, length, coeffs


def _gram_schmidt_classical_step(vec1, vec2):
    coeff = linalg.vecdot(vec1, vec2)
    vec_ortho = vec1 - coeff * vec2
    return vec_ortho, coeff


def _normalise(vec):
    length = linalg.vector_norm(vec)
    return vec / length, length


def _decompose_fori_loop(v0, *parameters, algorithm: _LanczosAlg):
    r"""Decompose a matrix purely based on matvec-products with A.

    The behaviour of this function is equivalent to

    ```python
    def decompose(v0, *matvec_funs, algorithm):
        init, step, extract, (lower, upper) = algorithm
        state = init(v0)
        for _ in range(lower, upper):
            state = step(state, *matvec_funs)
        return extract(state)
    ```

    but the implementation uses JAX' fori_loop.
    """
    init, step, extract, (lower, upper) = algorithm
    init_val = init(v0)

    def body_fun(_, s):
        return step(s, *parameters)

    result = control_flow.fori_loop(lower, upper, body_fun=body_fun, init_val=init_val)
    return extract(result)



"""Types."""

# fmt: off
from collections.abc import Callable  # noqa: F401
from typing import Any, Generic, Iterable, Sequence, Tuple, TypeVar  # noqa: F401, UP035

from jax import Array  # noqa: F401

# fmt: on



"""PyTree utilities."""

import jax
import jax.flatten_util


def tree_all(tree, /):
    return jax.tree_util.tree_all(tree)


def tree_map(func, tree, *rest):
    return jax.tree_util.tree_map(func, tree, *rest)


def ravel_pytree(tree, /):
    return jax.flatten_util.ravel_pytree(tree)


def tree_leaves(tree, /):
    return jax.tree_util.tree_leaves(tree)


def tree_structure(tree, /):
    return jax.tree_util.tree_structure(tree)


def partial_pytree(func, /):
    return jax.tree_util.Partial(func)



"""Tests for pseudo-inverse functionality."""

from matfree import pinv
from matfree.backend import func, linalg, np, prng, testing


def fun_tall(x):
    """Evaluate a nonlinear function with a tall Jacobian."""
    P = prng.uniform(prng.prng_key(seed=2), shape=(len(x) - 2, len(x)))
    return np.cos(P @ np.sin(x))


def fun_wide(x):
    """Evaluate a nonlinear function with a wide Jacobian."""
    P = prng.uniform(prng.prng_key(seed=2), shape=(len(x) + 2, len(x)))
    return np.cos(P @ np.sin(x))


def densify(fun, input_dim):
    """Densify a linear operator."""
    return func.vmap(fun, in_axes=-1, out_axes=-1)(np.eye(input_dim))


def solve_dense(Av, b):
    """Solve a linear system by densifying + LU decomposition."""
    return linalg.solve(func.jacfwd(Av)(b), b)


def solve_cg(Av, b):
    """Solve a linear system with conjugate gradients."""
    return linalg.cg(Av, b)[0]


@testing.parametrize(
    "fun, pinv_fun", ([fun_tall, pinv.pinv_tall], [fun_wide, pinv.pinv_wide])
)
@testing.parametrize("solve_fun", [solve_dense, solve_cg])
def test_inverse_correct(fun, pinv_fun, solve_fun):
    """Compare the values of the computed pseudo-inverses to numpy.pinv()."""
    key = prng.prng_key(seed=1)
    x = prng.uniform(key, shape=(7,))

    J = func.jacfwd(fun)(x)

    # Invert
    inverse = pinv_fun(lambda v: J @ v, lambda v: v @ J, solve=solve_fun)

    # Dense invert
    inverse_matrix = densify(inverse, len(fun(x)))
    inverse_expected = linalg.pinv(func.jacfwd(fun)(x))
    assert np.allclose(inverse_matrix, inverse_expected, rtol=1e-3, atol=1e-3)



"""Test the Golub-Kahan-Lanczos bi-diagonalisation with full re-orthogonalisation."""

from matfree import lanczos, test_util
from matfree.backend import linalg, np, prng, testing


@testing.fixture()
def A(nrows, ncols, num_significant_singular_vals):
    """Make a positive definite matrix with certain spectrum."""
    # 'Invent' a spectrum. Use the number of pre-defined eigenvalues.
    n = min(nrows, ncols)
    d = np.arange(n) + 10.0
    d = d.at[num_significant_singular_vals:].set(0.001)
    return test_util.asymmetric_matrix_from_singular_values(d, nrows=nrows, ncols=ncols)


@testing.parametrize("nrows", [50])
@testing.parametrize("ncols", [49])
@testing.parametrize("num_significant_singular_vals", [4])
@testing.parametrize("order", [6])  # ~1.5 * num_significant_eigvals
def test_lanczos_bidiag_full_reortho(A, order):
    """Test that Lanczos tridiagonalisation yields an orthogonal-tridiagonal decomp."""
    nrows, ncols = np.shape(A)
    key = prng.prng_key(1)
    v0 = prng.normal(key, shape=(ncols,))

    def Av(v):
        return A @ v

    def vA(v):
        return v @ A

    algorithm = lanczos.alg_bidiag_full_reortho(Av, vA, order, matrix_shape=np.shape(A))
    v0 /= linalg.vector_norm(v0)
    Us, Bs, Vs, (b, v) = algorithm(v0)
    (d_m, e_m) = Bs

    tols_decomp = {"atol": 1e-5, "rtol": 1e-5}

    assert np.shape(Us) == (nrows, order + 1)
    assert np.allclose(Us.T @ Us, np.eye(order + 1), **tols_decomp), Us.T @ Us

    assert np.shape(Vs) == (order + 1, ncols)
    assert np.allclose(Vs @ Vs.T, np.eye(order + 1), **tols_decomp), Vs @ Vs.T

    UAVt = Us.T @ A @ Vs.T
    assert np.allclose(linalg.diagonal(UAVt), d_m, **tols_decomp)
    assert np.allclose(linalg.diagonal(UAVt, 1), e_m, **tols_decomp)

    B = _bidiagonal_dense(d_m, e_m)
    assert np.shape(B) == (order + 1, order + 1)
    assert np.allclose(UAVt, B, **tols_decomp)

    em = np.eye(order + 1)[:, -1]
    AVt = A @ Vs.T
    UtB = Us @ B
    AtUt = A.T @ Us
    VtBtb_plus_bve = Vs.T @ B.T + b * v[:, None] @ em[None, :]
    assert np.allclose(AVt, UtB, **tols_decomp)
    assert np.allclose(AtUt, VtBtb_plus_bve, **tols_decomp)


def _bidiagonal_dense(d, e):
    diag = linalg.diagonal_matrix(d)
    offdiag = linalg.diagonal_matrix(e, 1)
    return diag + offdiag


@testing.parametrize("nrows", [5])
@testing.parametrize("ncols", [3])
@testing.parametrize("num_significant_singular_vals", [3])
def test_error_too_high_depth(A):
    """Assert that a ValueError is raised when the depth exceeds the matrix size."""
    nrows, ncols = np.shape(A)
    max_depth = min(nrows, ncols) - 1

    with testing.raises(ValueError):

        def eye(v):
            return v

        _ = lanczos.alg_bidiag_full_reortho(
            eye, eye, max_depth + 1, matrix_shape=np.shape(A)
        )


@testing.parametrize("nrows", [5])
@testing.parametrize("ncols", [3])
@testing.parametrize("num_significant_singular_vals", [3])
def test_error_too_low_depth(A):
    """Assert that a ValueError is raised when the depth is negative."""
    min_depth = 0
    with testing.raises(ValueError):

        def eye(v):
            return v

        _ = lanczos.alg_bidiag_full_reortho(
            eye, eye, min_depth - 1, matrix_shape=np.shape(A)
        )


@testing.parametrize("nrows", [15])
@testing.parametrize("ncols", [3])
@testing.parametrize("num_significant_singular_vals", [3])
def test_no_error_zero_depth(A):
    """Assert the corner case of zero-depth does not raise an error."""
    nrows, ncols = np.shape(A)
    key = prng.prng_key(1)
    v0 = prng.normal(key, shape=(ncols,))

    def Av(v):
        return A @ v

    def vA(v):
        return v @ A

    algorithm = lanczos.alg_bidiag_full_reortho(Av, vA, 0, matrix_shape=np.shape(A))
    Us, Bs, Vs, (b, v) = algorithm(v0)
    (d_m, e_m) = Bs
    assert np.shape(Us) == (nrows, 1)
    assert np.shape(Vs) == (1, ncols)
    assert np.shape(d_m) == (1,)
    assert np.shape(e_m) == (0,)
    assert np.shape(b) == ()
    assert np.shape(v) == (ncols,)


@testing.parametrize("nrows", [15])
@testing.parametrize("ncols", [3])
@testing.parametrize("num_significant_singular_vals", [3])
@testing.parametrize("order", [2])
def test_validate_unit_norm(A, order):
    """Test that the outputs are NaN if the input is not normalized."""
    nrows, ncols = np.shape(A)
    key = prng.prng_key(1)

    # Not normalized!
    v0 = prng.normal(key, shape=(ncols,)) + 1.0

    def Av(v):
        return A @ v

    def vA(v):
        return v @ A

    algorithm = lanczos.alg_bidiag_full_reortho(
        Av, vA, order, matrix_shape=np.shape(A), validate_unit_2_norm=True
    )
    Us, (d_m, e_m), Vs, (b, v) = algorithm(v0)

    # Since v0 is not normalized, all inputs are NaN
    for x in (Us, d_m, e_m, Vs, b, v):
        assert np.all(np.isnan(x))
